"""Classes and functions related to the interferometry lasers."""

from functools import partial

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import Colormap

from . import convert
from .zern import FIRST_INDEX_J, ZernikeNorm, ZernikeOrder, ZernikePolynomial


class Wavevectors:
    """
    Class that defines the wave vectors of the two Ramen beams.

    Parameters
    ----------
    k1, k2 : float
        1D wave vectors (wavenumber) in z-direction of the two Raman beams  in rad/m,
        defaults to 2*pi/780e-9

    Attributes
    ----------
    k1, k2 : float
        1D wave vectors (wavenumber) in z-direction of the two Raman beams in rad/m
    """

    def __init__(self, k1=8055366, k2=-8055366):
        self.k1 = k1
        self.k2 = k2

    def doppler_shift(self, atoms):
        """
        Calculate the Doppler shifts for an atomic ensemble.

        Parameters
        ----------
        atoms : AtomicEnsemble
            an atomic enemble with a finite velocity in the z direction

        Returns
        -------
        dopler_shift : ndarray
            Doppler shift of each atom in the ensemble in rad/s
        """
        # calculate two photon detuning for atoms' velocity (-v*k_eff)
        velocity_z = atoms.velocity[:, 2]
        doppler_shift = -velocity_z * (self.k1 - self.k2)
        return doppler_shift


class IntensityProfile:
    """
    Class that defines a Gaussian intensity profile in terms of the Rabi frequency.

    Parameters
    ----------
    r_profile : float
        1/e² radius of the Gaussian intensity profile in m
    center_rabi_freq : float
        Rabi frequency at center of intensity profile in rad/s
    r_beam : float (optional)
        Beam radius in m. Can be set if the intensity profile is limited by an aperture.
        Rabi frequency will be set to 0 outside of the beam.

    Attributes
    ----------
    r_beam : float or None
        Beam radius in m. If set, Rabi frequency will be set to 0 outside of the beam.
    profile_func : function
        Function that calculates the Rabi frequency at a position of a Gaussian beam.
    """

    def __init__(self, *, r_profile, center_rabi_freq, r_beam=None):
        self.r_beam = r_beam
        self.profile_func = partial(
            self._gaussian_profile,
            r_profile=r_profile,
            center_rabi_freq=center_rabi_freq,
        )

    def get_rabi_freq(self, pos):
        """
        Rabi frequency at a position of a Gaussian beam.

        Parameters
        ----------
        pos : (n x 2) array or (n x 3) array
            positions of the n atoms in two or three dimensions (x, y, [z]).

        Returns
        -------
        rabi_freqs : array of float
            the Rabi frequencies for the n positions. If `r_beam` is set, the Rabi
            frequency will be set to 0 outside of the beam.
        """
        rabi_freqs = self.profile_func(pos)
        if self.r_beam is not None:
            rabi_freqs[pos[:, 0] ** 2 + pos[:, 1] ** 2 > self.r_beam] = 0.0
        return rabi_freqs

    @staticmethod
    def _gaussian_profile(pos, r_profile: float, center_rabi_freq: float):
        """
        Rabi frequency at a position of a Gaussian beam.

        Parameters
        ----------
        pos : (n x 2) array or (n x 3) array
            positions of the n atoms in two or three dimensions (x, y, [z]).
        r_profile : float
            1/e² radius of the Gaussian intensity profile in m
        center_rabi_freq : float
            Rabi frequency at center of intensity profile in rad/s

        Returns
        -------
        rabi_freqs : array of float
            the Rabi frequencies for the n positions. If `r_beam` is set, the Rabi
            frequency will be set to 0 outside of the beam.
        """
        r_sq = pos[:, 0] ** 2 + pos[:, 1] ** 2
        rabi_freqs = center_rabi_freq * np.exp(-2 * r_sq / (r_profile**2))
        return rabi_freqs


class Wavefront:
    """
    Class that defines a wavefront.

    Parameters
    ----------
    r_wf : float
        radius of the wavefront data in m
    coeff : dict
        Dictionary of the Zernike coefficients. The keys are the Zernike polynomial
        single indices j and the values are the coefficients depends on the Zernike
        order and normalization.
    r_beam : float (optional)
        Beam radius in m. Can be set if the beam is smaller than the wavefront data.
        Values outside of the beam will be set to NaN.
    zern_order : ZernikeOrder or str
        Ordering scheme for the Zernike polynomials. See `ZernikeOrder` for possible
        values.
    zern_norm : ZernikeNorm, str or None
        Normalization scheme for the Zernike polynomials. See `ZernikeNorm` for possible
        values.

    Attributes
    ----------
    r_wf : float
        radius of the wavefront data in m
    coeff : dict
        Dictionary of the Zernike coefficients. The keys are the Zernike polynomial
        single indices j and the values are the coefficients depends on the Zernike
        order and normalization.
    r_beam : float or None
        Beam radius in m. If set, values outside of the beam will be set to NaN.
    zern_order : ZernikeOrder or str
        Ordering scheme for the Zernike polynomials.
    zern_norm : ZernikeNorm, str or None
        Normalization scheme for the Zernike polynomials.
    """

    def __init__(
        self,
        r_wf: float,
        coeff: dict[int, float],
        r_beam: float | None = None,
        zern_order: ZernikeOrder = ZernikeOrder.WYANT,
        zern_norm: ZernikeNorm | None = None,
    ) -> None:
        self.r_wf = r_wf
        self.coeff = coeff
        self.r_beam = r_beam
        self.zern_order = zern_order
        self.zern_norm = zern_norm

    def get_value(self, pos: np.ndarray) -> np.ndarray:
        """
        Get the wavefront at a position.

        Parameters
        ----------
        pos : n x 3 array
            array of position vectors (x, y, z) where the wavefront is probed

        Returns
        -------
        wf : nd array
            The value of the wavefront at the positions
        """
        pos = convert.cart2pol(pos)
        rho = pos[:, 0]
        rho = rho / self.r_wf
        rho[rho > 1] = np.nan
        theta = pos[:, 1]
        zern = ZernikePolynomial(self.coeff, self.zern_order, self.zern_norm)
        values = zern.zern_sum(rho, theta)
        if self.r_beam is not None:
            values[rho > self.r_beam / self.r_wf] = np.nan
        return values

    def plot(
        self,
        ax: plt.Axes | None = None,
        cmap: str | Colormap = "RdBu",
        levels: int = 100,
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """
        Plot the wavefront data.

        Parameters
        ----------
        ax : Axis , optional
            If axis is provided, they will be used for the plot. if not provided, a new
            plot will automatically be created.
        cmap : str or Colormap
            Colormap for the plot
        level : int
            Number of levels for the contour plot
        **kwargs
            Additional keyword arguments for the plot function

        Returns
        -------
        fig, ax : tuple of plt.Figure and plt.Axes
            The figure and axis of the plot
        """
        azimuths = np.radians(np.linspace(0, 360, 180))
        zeniths = np.linspace(0, self.r_wf, 50)
        rho, theta = np.meshgrid(zeniths, azimuths)
        z = np.zeros_like(rho)

        n_dim, m_dim = rho.shape
        pos = np.array([rho.flatten(), theta.flatten(), z.flatten()]).T
        values = self.get_value(convert.pol2cart(pos))

        if ax is None:
            fig, ax = plt.subplots(subplot_kw=dict(projection="polar"))
        else:
            fig = ax.figure

        theta = theta.reshape(n_dim, m_dim)
        rho = rho.reshape(n_dim, m_dim)
        values = values.reshape(n_dim, m_dim)
        contour = ax.contourf(theta, rho, values, cmap=cmap, levels=levels, **kwargs)
        cbar = plt.colorbar(contour)
        cbar.set_label(r"Aberration / $\lambda$", rotation=90)
        plt.tight_layout()

        return fig, ax

    def plot_coeff(
        self, ax: plt.Axes | None = None, **kwargs
    ) -> tuple[plt.Figure, plt.Axes]:
        """
        Plot the coefficients as a bar chart.

        Parameters
        ----------
        ax : Axis , optional
            If axis is provided, they will be used for the plot. if not provided, a new
            plot will automatically be created.

        Returns
        -------
        fig, ax : tuple of plt.Figure and plt.Axes
            The figure and axis of the plot
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        ax.bar(list(self.coeff.keys()), list(self.coeff.values()), **kwargs)
        ax.set_xlabel("Zernike polynomial $j$")
        ax.set_ylabel(r"Zernike coefficient $Z_j$ / $\lambda$")
        ax.set_xlim(min(self.coeff.keys()) - 1, max(self.coeff.keys()) + 1)
        return fig, ax


def gen_wavefront(
    r_wf: float,
    std: float = 0.0,
    r_beam: float | None = None,
    n_zern: int = 36,
    zern_order: ZernikeOrder = ZernikeOrder.NOLL,
    zern_norm: ZernikeNorm | None = None,
    seed: int | None = None,
) -> Wavefront:
    """
    Create an artificial wavefront.

    Parameters
    ----------
    r_wf : float
        radius of the wavefront data in m
    std : float
        standard deviation of each Zernike polynomial coefficient in multiples of the
        wavelength.
    r_beam : float, optional
        Beam radius in m. Can be set if the beam is smaller than the wavefront data.
        Values outside of the beam will be set to NaN.
    n_zern : int
        number of Zernike polynomials to be used
    zern_order : ZernikeOrder or str
        Ordering scheme for the Zernike polynomials. See `ZernikeOrder` for possible
        values.
    zern_norm : ZernikeNorm, str or None
        Normalization scheme for the Zernike polynomials. See `ZernikeNorm` for possible
        values.
    seed : int, optional
        seed for the random number generator

    Returns
    -------
    wf : Wavefront
        artificial wavefront
    """
    if seed is not None:
        np.random.seed(seed)
    coeff = np.random.normal(0, std, size=n_zern)
    zern_coeff = {
        j: val for j, val in enumerate(coeff, start=FIRST_INDEX_J[zern_order])
    }
    return Wavefront(r_wf, zern_coeff, r_beam, zern_order, zern_norm)



"""Module for functions related to atomic position and velocity distributions."""

import numpy as np
from scipy.interpolate import interp1d

from .convert import vel_from_temp


def position_dist_gaussian(n: int, std: float) -> np.ndarray:
    """Create a random position distribution from a Gaussian distribution.

    Parameters
    ----------
    n : int
        number of samples
    std : float
        standard deviation of the Gaussian distribution in meters
    """
    return np.random.normal(scale=std, size=n)


def velocity_dist_from_temp(n: int, temperature: float) -> np.ndarray:
    """Create a random velocity distribution from a given temperature.

    Parameters
    ----------
    n : int
        number of samples
    temperature : float
        temperature of the atomic ensemble in Kelvin

    Returns
    -------
    velocities : array
        n-dimensional array of the randomly selected velocities
    """
    return np.random.normal(scale=vel_from_temp(temperature), size=n)


def velocity_dist_for_box_pulse_velsel(
    n: int, pulse_duration: float, wavelenth: float = 780e-9, n_lobes: int = 3
) -> np.ndarray:
    """Create a random velocity distribution for a box pulse velocity selection.

    The velocity distribution is created by sampling the inverse cumulative distribution
    function of the Fourier transform of a box pulse with a uniform distribution. The
    box pulse is defined by its duration and the velocity distribution is given by
    |τ sinc(2*τ*v/λ)|^2 where τ is the pulse duration, λ the wavelength and v is the
    velocity of the atoms.

    Parameters
    ----------
    n : int
        number of samples
    pulse_duration : float
        duration of the box pulse in seconds
    wavelenth : float, optional
        wavelength of the light in meters (default 780e-9)
    n_lobes : int, optional
        number of lobes of the sinc function that are sampled (default 3)

    Returns
    -------
    velocities : array
        n-dimensional array of the randomly selected velocities
    """

    def box_pulse_ft(x: np.ndarray, pulse_duration: float) -> np.ndarray:
        """Fourier transform of a box pulse."""
        return np.abs(pulse_duration * np.sinc(pulse_duration * x) ** 2)

    # Based on https://stackoverflow.com/a/64288861/2750945
    x = np.linspace(-n_lobes / pulse_duration, n_lobes / pulse_duration, int(1e5))
    y_cumsum = np.cumsum(box_pulse_ft(x, pulse_duration))
    y_cumsum = y_cumsum - y_cumsum.min()
    f = interp1d(y_cumsum / y_cumsum.max(), x)

    return wavelenth * f(np.random.random(n)) / 2


def velocity_dist_for_gaussian_velsel(
    n: int, pulse_duration: float, wavelength: float = 780e-9
):
    """Create a random velocity distribution for a Gaussian pulse velocity selection.

    Parameters
    ----------
    n : int
        number of samples
    pulse_duration : float
        duration of the Gaussian pulse in seconds (1/sqrt(2) full width, i.e. 2*sigma)
    wavelength : float, optional
        wavelength of the light in meters (default 780e-9)

    Returns
    -------
    velocities : array
        n-dimensional array of the randomly selected velocities
    """

    pulse_duration = pulse_duration / 2  # Convert to sigma
    sigma_nu = 1 / (2 * np.pi * pulse_duration)

    return np.random.normal(scale=sigma_nu, size=n) * wavelength / 2



"""Classes and functions related to the atomic cloud."""

from functools import partial
from typing import Callable, Literal

import matplotlib.pyplot as plt
import numpy as np
import scipy.linalg as splin

from .dist import position_dist_gaussian, velocity_dist_from_temp


class AtomicEnsemble:
    """
    Represents an atomic ensemble consisting of n atoms.

    Each atom is is defined by its phase space vector (x0, y0, z0, vx, vy, vz) at time
    t=0. From this phase space vector the position at later times can be calculated.

    Parameters
    ----------
    phase_space_vectors : ndarray
        n x 6 dimensional array representing the phase space vectors
        (x0, y0, z0, vx, vy, vz) of the atoms in an atomic ensemble
    state_kets : m x 1 or n x m x 1 array or list, optional
        vector(s) representing the `m` internal degrees of freedom of the atoms. If the
        list or array is one-dimensional, all atoms are initialized in the same internal
        state. Alternatively, each atom can be initialized with a different state vector
        by passing an array of state vectors for every atom. E.g. to initialize all
        atoms in the ground state of a two-level system, pass `[1, 0]` which is the
        default.
    time : float, optional
        the initial time (default 0) when the phase space and state vectors are
        initialized

    Attributes
    ----------
    phase_space_vectors : ndarray
        n x 6 dimensional array representing the phase space vectors
        (x0, y0, z0, vx, vy, vz) of the atoms in an atomic ensemble
    """

    def __init__(self, phase_space_vectors, state_kets=[1, 0], time: float = 0.0):
        assert phase_space_vectors.shape[1] == 6
        self.phase_space_vectors = phase_space_vectors
        self.state_kets = state_kets
        self.time = time
        # for the future when we might implement forces
        self.initial_position = self.phase_space_vectors[:, 0:3]
        self.initial_velocity = self.phase_space_vectors[:, 3:6]

    def __getitem__(self, key):
        """Select  certain atoms "from the ensemble.

        Parameters
        ----------
            key : int or slice or bool map
                for example 2, 1:15 or a boolean map

            Returns
            -------
            new_instance : AtomicEnsemble
                a new instance of the atomic ensemble only containing the selected atoms
        """
        phase_space_vectors = self.phase_space_vectors[key][:]
        state_kets = self.state_kets[key][:]
        if isinstance(key, int):
            # retain correct shape in case of only one atom is selected
            phase_space_vectors = phase_space_vectors.reshape(1, 6)
            state_kets = state_kets.reshape(1, len(state_kets))
        new_instance = AtomicEnsemble(
            phase_space_vectors=phase_space_vectors,
            state_kets=state_kets,
        )
        return new_instance

    def __len__(self):
        """Return the number of atoms in the ensemble."""
        return self.phase_space_vectors.shape[0]

    @property
    def time(self):
        """float: Time changes when propagating the atomic ensemble."""
        return self._time

    @time.setter
    def time(self, value):
        self._time = value

    @property
    def state_kets(self):
        """(n x m x 1) array: The ket vectors of the m level system."""
        return self._state_kets

    @state_kets.setter
    def state_kets(self, new_state_kets):
        if isinstance(new_state_kets, list):
            new_state_kets = np.array([new_state_kets]).T
        if new_state_kets.ndim == 2:
            # state vector is the same for all atoms
            self._state_kets = np.repeat([new_state_kets], len(self), axis=0)
        else:
            # there has to be a state vector for every atom in the ensemble
            assert new_state_kets.shape[0] == len(self)
            self._state_kets = new_state_kets

    @property
    def state_bras(self):
        """(n x 1 x m) array: The bra vectors of the m level system."""
        # exchange second and third index, then complex conjugate
        return np.conjugate(np.einsum("ijk->ikj", self.state_kets))

    @property
    def density_matrices(self):
        """
        (n x m x m) array: Density matrix of the m level system of the n atoms.

        These are pure states.
        """
        # |Psi><Psi|
        return np.einsum("ijk,ijk->ijk", self.state_kets, self.state_bras)

    @property
    def density_matrix(self):
        """(m x m) array: Density matrix of the ensemble's m level system."""
        pure_dm = self.density_matrices
        n_atoms = self.state_kets.shape[0]
        # sum over pure |Psi><Psi| and divide by N
        return 1 / n_atoms * np.einsum("ijk->jk", pure_dm)

    @property
    def position(self):
        """(n x 3) array: Positions (x, y, z) of the atoms in the ensemble."""
        return self.phase_space_vectors[:, 0:3]

    @position.setter
    def position(self, new_position):
        self.phase_space_vectors[:, 0:3] = new_position

    @property
    def velocity(self):
        """n x 3) array: Velocities (vx, vy, vz) of the atoms in the ensemble."""
        return self.phase_space_vectors[:, 3:6]

    @velocity.setter
    def velocity(self, new_velocity):
        self.phase_space_vectors[:, 3:6] = new_velocity

    def calc_position(self, t):
        """
        Calculate the positions (x, y, z) of the atoms after propagation.

        Parameters
        ----------
        t : float
            time when the positions should be calculated

        Returns
        -------
        pos : array
            n x 3 dimensional array of the positions (x, y, z)
        """
        return self.initial_position + self.initial_velocity * t

    def state_occupation(self, state):
        """
        Calculate the state population of each atom in the ensemble.

        Parameters
        ----------
        state : int or list_like
            Specifies which state population should be calculated. E.g. the excited
            state of a two-level system can be calculated by passing either 1 or [0, 1].

        Returns
        -------
        occupation : array
            n dimensional array of the state population of each of the n atom
        """
        # create bra on which the kets of the atomic ensemble are projected
        if isinstance(state, (int, np.integer)):
            # list of zeros
            zeros = [0] * self.state_kets.shape[1]
            # set entry to 1, overwrite state variable
            zeros[state] = 1
            state = zeros
        if isinstance(state, list):
            # check that bases match
            assert len(state) == self.state_kets.shape[1]
            state = np.array(state)
        projection_bras = np.repeat(np.array([[state]]), len(self), axis=0)
        # |<i|Psi>|^2
        occupation = np.abs(np.matmul(projection_bras, self.state_kets)) ** 2
        return occupation.flatten()

    def fidelity(self, rho_target):
        """
        Calculate fidelity of ensemble's density matrix and target matrix [1].

        Parameters
        ----------
        rho_target : array
            target density matrix as m x m array

        Returns
        -------
        fidelity : float
            fidelity of AtomicEnsemble's compared to target density matrix

        References
        ----------
        [1] https://en.wikipedia.org/wiki/Fidelity_of_quantum_states
        """
        return _fidelity(self.density_matrix, rho_target)

    def plot(
        self,
        ax: plt.Axes | None = None,
        view_from: Literal["x", "y", "z"] = "z",
        bins: int = 50,
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Plot the positions of the atoms in the ensemble.

        ax : Axis, optional
            If axis is provided, they will be used for the plot. if not provided, a new
            plot will automatically be created.
        view_from : str
            View from which direction the plot is created. Options are "x", "y", "z".
        bins : int
            Number of bins for the histogram
        **kwargs
            Additional keyword arguments for the plot function

        Returns
        -------
        fig, ax : tuple of plt.Figure and plt.Axes
            The figure and axis of the plot
        """
        if ax is None:
            fig, ax = plt.subplots(subplot_kw=dict(projection="polar"))
        else:
            fig = ax.figure

        view = {
            "x": (1, 2),
            "y": (0, 2),
            "z": (0, 1),
        }
        ax.hist2d(
            self.position[:, view[view_from][0]],
            self.position[:, view[view_from][1]],
            bins=bins,
            **kwargs,
        )

        return fig, ax


def create_random_ensemble(
    n_samples: int,
    *,
    mean_x: float = 0.0,
    mean_y: float = 0.0,
    mean_z: float = 0.0,
    mean_vx: float = 0.0,
    mean_vy: float = 0.0,
    mean_vz: float = 0.0,
    x_dist: Callable[[int], np.ndarray] = partial(position_dist_gaussian, std=1.5e-3),
    y_dist: Callable[[int], np.ndarray] = partial(position_dist_gaussian, std=1.5e-3),
    z_dist: Callable[[int], np.ndarray] = partial(position_dist_gaussian, std=1.5e-3),
    vx_dist: Callable[[int], np.ndarray] = partial(
        velocity_dist_from_temp, temperature=3e-6
    ),
    vy_dist: Callable[[int], np.ndarray] = partial(
        velocity_dist_from_temp, temperature=3e-6
    ),
    vz_dist: Callable[[int], np.ndarray] = partial(
        velocity_dist_from_temp, temperature=160e-9
    ),
    seed: int | None = None,
    **kwargs,
) -> AtomicEnsemble:
    """Create a random atomic ensemble from given distributions.

    Have a look at the example notebooks to see how to use `partial` to create
    distributions with different user defined parameters.

    Parameters
    ----------
    n_samples : int
        number of random samples
    mean_x, mean_y, mean_z : float, optional
        mean position of the atomic ensemble in meters (default 0.0)
    mean_vx, mean_vy, mean_vz : float, optional
        mean velocity of the atomic ensemble in meters per second (default 0.0)
    x_dist, y_dist, z_dist : callable
        Function that returns a random position distribution. The function takes a
        single integer argument `n` and returns a n-dimensional array of random samples.
        To be used with `partial` to set the parameters of the distributions from the
        `dist` module.
    vx_dist, vy_dist, vz_dist : callable
        Function that returns a random velocity distribution. The function takes a
        single integer argument `n` and returns a n-dimensional array of random samples.
        To be used with `partial` to set the parameters of the distributions from the
        `dist` module.
    seed : int or 1-d array_like, optional
        Set the seed of the random number generator to get predictable samples. If set,
        this number is passed to `numpy.random.seed`.
    **kwargs :
        Optional keyworded arguments passed to `AtomicEnsemble`

    Returns
    -------
    ensemble : AtomicEnsemble
        Atomic ensemble containing the generated phase space vectors.
    """

    if seed is not None:
        np.random.seed(seed)

    # initialize vector with phase-space entries and fill them
    phase_space_vectors = np.zeros((n_samples, 6))
    phase_space_vectors[:, 0] = x_dist(n_samples) + mean_x
    phase_space_vectors[:, 1] = y_dist(n_samples) + mean_y
    phase_space_vectors[:, 2] = z_dist(n_samples) + mean_z
    phase_space_vectors[:, 3] = vx_dist(n_samples) + mean_vx
    phase_space_vectors[:, 4] = vy_dist(n_samples) + mean_vy
    phase_space_vectors[:, 5] = vz_dist(n_samples) + mean_vz
    ensemble = AtomicEnsemble(phase_space_vectors, **kwargs)
    return ensemble


def _fidelity(rho_a, rho_b):
    """
    Calculate the fidelity of two density matrices [1, 2].

    Parameters
    ----------
    rho_a : array
        density matrix as m x m array
    rho_b : array
        density matrix as m x m array

    Returns
    -------
    fidelity : float
        fidelity of both density matrices

    References
    ----------
    [1] https://en.wikipedia.org/wiki/Fidelity_of_quantum_states
    [2] http://qutip.org/docs/4.0.2/modules/qutip/metrics.html
    """
    assert rho_a.shape == rho_b.shape
    sqrt_rho_a = splin.sqrtm(rho_a)  # matrix square root
    # Implementation used in qutip"s fidelity calculation: calculating the eigenvalues
    # and taking it's square root instead of matrix square root and taking the trace.
    # It's faster.
    eig_vals = np.linalg.eigvals(sqrt_rho_a @ rho_b @ sqrt_rho_a)
    fidelity = np.real(np.sum(np.sqrt(eig_vals))) ** 2
    return fidelity



import json
from functools import partial

import numpy as np

import aisim as ais


def test_wf():
    data = json.load(open("docs/examples/data/experimental_data.json"))
    r_dets = np.array(data["r_det"])
    grav_data = np.array(data["g"])

    coeff_window = {
        j: val for j, val in enumerate(np.loadtxt("docs/examples/data/wavefront.txt"))
    }
    wf = ais.Wavefront(
        10.91e-3, coeff_window, zern_norm=None, zern_order=ais.ZernikeOrder.WYANT
    )
    for n in [0, 1, 2]:
        wf.coeff[n] = 0  # remove piston, tip and tilt

    atoms = ais.create_random_ensemble(
        int(1e5),
        mean_x=0.0,
        mean_y=0.0,
        mean_z=0.0,
        mean_vx=0.0,
        mean_vy=0.0,
        mean_vz=0.0,
        x_dist=partial(ais.dist.position_dist_gaussian, std=3.0e-3),
        y_dist=partial(ais.dist.position_dist_gaussian, std=3.0e-3),
        z_dist=partial(ais.dist.position_dist_gaussian, std=0.0),
        vx_dist=partial(ais.dist.velocity_dist_from_temp, temperature=3.5e-6),
        vy_dist=partial(ais.dist.velocity_dist_from_temp, temperature=3.5e-6),
        vz_dist=partial(
            ais.dist.velocity_dist_for_gaussian_velsel, pulse_duration=60e-6
        ),
        state_kets=[0, 1],
        seed=1,
    )

    t_det = 778e-3  # time of the detection in s

    T = 260e-3  # interferometer time in s
    t1 = 130e-3  # time of first pulse in s
    t2 = t1 + T
    t3 = t2 + T

    awfs = []
    for r_det in r_dets:
        # creating detector with new detection radius
        det = ais.PolarDetector(t_det, r_det=r_det)

        det_atoms = det.detected_atoms(atoms)

        # calculate the imprinted phase for each "test atom" at each pulse. This is the
        # computationally heavy part
        phi1 = 2 * np.pi * wf.get_value(det_atoms.calc_position(t1))
        phi2 = 2 * np.pi * wf.get_value(det_atoms.calc_position(t2))
        phi3 = 2 * np.pi * wf.get_value(det_atoms.calc_position(t3))

        # calculate a complex amplitude factor for the Mach-Zehnder sequence and
        # weight their contribution to the signal
        awfs.append(np.nanmean(np.exp(1j * (phi1 - 2 * phi2 + phi3))))

    grav = 2 * ais.phase_error_to_grav(np.angle(awfs), T=260e-3, keff=1.610574779769e7)

    # test that simulation and data fit reasonably
    assert np.sqrt(np.sum((grav - grav_data) ** 2)) < 1.2e-8



from functools import partial

import numpy as np
import pytest  # noqa

import aisim as ais


def generate_random_atoms(n_atom, n_int):
    random_phase_space_vectors = np.random.rand(n_atom, 6)
    random_kets = np.random.rand(n_atom, n_int)
    # normalize random ket
    norm = np.einsum("ij,ij->i", random_kets, random_kets)
    norm_random_kets = np.einsum("ij,i->ij", random_kets, 1 / norm ** (1 / 2))
    # apply random phase shifts
    random_phases = 2 * np.pi * np.random.rand(n_atom, n_int)
    norm_random_kets = np.exp(1j * random_phases) * norm_random_kets
    norm_random_kets = np.reshape(norm_random_kets, (n_atom, n_int, 1))
    random_atoms = ais.AtomicEnsemble(random_phase_space_vectors, norm_random_kets)
    return random_atoms


def test_atomic_ensemble_methods():
    # This tests the AtomicEnsemble class' methods for consistency. It creates ensembles
    # of atoms with random internal and external states.
    # Methods of AtomicEnsemble that are tested:
    #     state_kets, state_bras, density_matrices, density_matrix
    def atomic_ensemble_test_function(atoms):
        # Testing AtomicEnsemble properties
        assert atoms.state_kets.shape == (n_atom, n_int, 1)
        assert atoms.state_bras.shape == (n_atom, 1, n_int)
        assert atoms.phase_space_vectors.shape == (n_atom, 6)
        assert atoms.phase_space_vectors.shape == (n_atom, 6)
        assert (atoms.phase_space_vectors[:, 0:3] == atoms.position).all()
        assert (atoms.calc_position(0) == atoms.position).all()
        assert (atoms.phase_space_vectors[:, 3:6] == atoms.velocity).all()
        # Test duality of Kets and Bras
        np.testing.assert_array_almost_equal(
            atoms.state_kets, np.conjugate(np.transpose(atoms.state_bras, (0, 2, 1)))
        )
        # Test whether the trace of every density matrix is equal to one
        np.testing.assert_almost_equal(
            np.trace(atoms.density_matrices, axis1=1, axis2=2), 1
        )
        # Test whether every single atoms density matrix satisfiesrho^2 = rho (condition
        # for pure states)
        np.testing.assert_array_almost_equal(
            np.matmul(atoms.density_matrices, atoms.density_matrices),
            atoms.density_matrices,
        )
        # test for the density matrix (mixed states)
        np.testing.assert_almost_equal(np.trace(atoms.density_matrix), 1)
        # test whether the sum of all state occupation is 1
        m_levels = np.arange(atoms.state_kets.shape[1])  # number of levels
        state_occupation = 0
        for m in m_levels:
            state_occupation += atoms.state_occupation(m)
        np.testing.assert_almost_equal(state_occupation, 1)
        # Test properties of the fidelity function
        # fidelity of density matrix with itself is 1
        np.testing.assert_almost_equal(1, atoms.fidelity(atoms.density_matrix))

    # Test AtomicEnsemble from very general randomly generated states
    n_atom = 1000
    n_ints = [1, 2, 3, 100]
    for n_int in n_ints:
        random_atoms = generate_random_atoms(n_atom, n_int)
        atomic_ensemble_test_function(random_atoms)

        # test slicing/selecting
        # selecting a single atom
        one_atom = random_atoms[0]
        assert len(one_atom) == 1
        if len(random_atoms) > 5:
            sliced_atoms = random_atoms[0:5]
            assert len(sliced_atoms) == 5

    # Create an ensemble of atoms
    atoms = ais.create_random_ensemble(
        n_atom,
        mean_x=1.0,
        mean_y=1.0,
        mean_z=1.0,
        x_dist=partial(ais.dist.position_dist_gaussian, std=1.0),
        y_dist=partial(ais.dist.position_dist_gaussian, std=1.0),
        z_dist=partial(ais.dist.position_dist_gaussian, std=1.0),
        vx_dist=partial(ais.dist.velocity_dist_from_temp, temperature=3.0e-6),
        vy_dist=partial(ais.dist.velocity_dist_from_temp, temperature=30.0e-6),
        vz_dist=partial(ais.dist.velocity_dist_from_temp, temperature=0.2e-6),
        seed=1,
        state_kets=random_atoms.state_kets,
    )

    atomic_ensemble_test_function(atoms)


def test_fidelity():
    for n_atom in [1, 3, 1000]:
        for n_int in [2, 3, 10, 50]:
            # Generate random density matrices
            rho_a = generate_random_atoms(n_atom, n_int).density_matrix
            rho_b = generate_random_atoms(n_atom, n_int).density_matrix
            # Test that return is float
            assert isinstance(ais.atoms._fidelity(rho_a, rho_b), float)
            # Test symmetry F(rhoA,rhoB) == F(rhoB,rhoA)
            np.testing.assert_almost_equal(
                ais.atoms._fidelity(rho_a, rho_b), ais.atoms._fidelity(rho_b, rho_a)
            )
            # Test if F(rhoA, rhoB) is equal or less than 1
            assert np.all(ais.atoms._fidelity(rho_a, rho_b) <= 1)
            # Test if F(rhoA, rhoA) is equal to 1
            np.testing.assert_array_almost_equal(ais.atoms._fidelity(rho_a, rho_a), 1)



import numpy as np
import pytest

from aisim.zern import ZernikeNorm, ZernikeOrder, j_to_n_m, zernike_term

# Explicit test cases from https://en.wikipedia.org/wiki/Zernike_polynomials
noll_mapping_examples = [
    (0, 0),
    (1, 1),
    (1, -1),
    (2, 0),
    (2, -2),
    (2, 2),
    (3, -1),
    (3, 1),
    (3, -3),
    (3, 3),
    (4, 0),
    (4, 2),
    (4, -2),
    (4, 4),
    (4, -4),
    (5, 1),
    (5, -1),
    (5, 3),
    (5, -3),
    (5, 5),
]

ansi_mapping_examples = [
    (0, 0),
    (1, -1),
    (1, 1),
    (2, -2),
    (2, 0),
    (2, 2),
    (3, -3),
    (3, -1),
    (3, 1),
    (3, 3),
    (4, -4),
    (4, -2),
    (4, 0),
    (4, 2),
    (4, 4),
    (5, -5),
    (5, -3),
    (5, -1),
    (5, 1),
    (5, 3),
]

wyant_mapping_examples = [
    (0, 0),
    (1, 1),
    (1, -1),
    (2, 0),
    (2, 2),
    (2, -2),
    (3, 1),
    (3, -1),
    (4, 0),
    (3, 3),
    (3, -3),
    (4, 2),
    (4, -2),
    (5, 1),
    (5, -1),
    (6, 0),
    (4, 4),
    (4, -4),
    (5, 3),
    (5, -3),
]


def test_j_to_n_m():
    for j, (n, m) in enumerate(noll_mapping_examples, start=1):
        assert j_to_n_m(j, ZernikeOrder.NOLL) == (n, m)

    for j, (n, m) in enumerate(ansi_mapping_examples):
        assert j_to_n_m(j, ZernikeOrder.ANSI) == (n, m)

    for j, (n, m) in enumerate(wyant_mapping_examples):
        assert j_to_n_m(j, ZernikeOrder.WYANT) == (n, m)

    for j, (n, m) in enumerate(wyant_mapping_examples, start=1):
        assert j_to_n_m(j, ZernikeOrder.FRINGE) == (n, m)

    with pytest.raises(ValueError):
        j_to_n_m(-1, "unknown")

    with pytest.raises(ValueError):
        j_to_n_m(10000, ZernikeOrder.WYANT)

    with pytest.raises(ValueError):
        j_to_n_m(0, ZernikeOrder.FRINGE)

    with pytest.raises(ValueError):
        j_to_n_m(-1, ZernikeOrder.ANSI)


def test_normalization():
    # Test for Example p. 110 of https://www.thorlabs.com/drawings/8b3619460cf44a9-01039B9C-E2C2-1948-D47C301CE6F71CB0/WFS40-5C-Manual.pdf
    defocus = -1e-6 * zernike_term(np.array([0, 1]), 0, n=2, m=0, norm=ZernikeNorm.OSA)
    assert np.round(1e6 * (defocus[0] - defocus[1]), 2) == 3.46



from functools import partial

import matplotlib.pyplot as plt
import numpy as np

import aisim as ais


def test_wavefront():
    # test the methods of Wavefront that are not covered by wf_test.py
    wf = ais.gen_wavefront(2, r_beam=1.0, seed=0)
    wf.plot()
    wf.plot_coeff()
    _, ax = plt.subplots()
    wf.plot(ax=ax)
    _, ax = plt.subplots()
    wf.plot_coeff(ax=ax)

    pos = np.array([[0, 0, 0], [1.1, 0, 0], [0, 2.1, 0]])
    res2 = wf.get_value(pos)
    assert not np.isnan(res2[0])
    assert np.isnan(res2[1])
    assert np.isnan(res2[2])


def test_wavevectors():
    wv1 = ais.Wavevectors(1, -1)
    wv2 = ais.Wavevectors(-1, 1)

    atoms = ais.create_random_ensemble(
        10,
        mean_x=1.0,
        mean_y=1.0,
        mean_z=1.0,
        mean_vx=1.0,
        mean_vy=1.0,
        mean_vz=1.0,
        x_dist=partial(ais.dist.position_dist_gaussian, std=1.0),
        y_dist=partial(ais.dist.position_dist_gaussian, std=1.0),
        z_dist=partial(ais.dist.position_dist_gaussian, std=3.0e-3),
        vx_dist=partial(ais.dist.velocity_dist_from_temp, temperature=3.0e-6),
        vy_dist=partial(ais.dist.velocity_dist_from_temp, temperature=3.0e-6),
        vz_dist=partial(
            ais.dist.velocity_dist_for_box_pulse_velsel, pulse_duration=100e-6
        ),
        seed=0,
    )

    diff = wv1.doppler_shift(atoms) + wv2.doppler_shift(atoms)
    assert diff.all() == 0


def test_intensity_profile():
    r_profile = 1
    pos1 = np.array([[0, 0, 0]])
    pos2 = np.array([[0, 1, 0]])
    intensity_profile = ais.IntensityProfile(r_profile=r_profile, center_rabi_freq=1)

    rabi1 = intensity_profile.get_rabi_freq(pos1)
    rabi2 = intensity_profile.get_rabi_freq(pos2)
    ratio = rabi2 / rabi1
    # test that Rabi frequency has dropped to 1/e**2 of center value
    np.testing.assert_almost_equal(ratio[0], 1 / np.e**2)

    # test that Rabi frequency is zero outside of the beam
    intensity_profile2 = ais.IntensityProfile(
        r_profile=r_profile, center_rabi_freq=1, r_beam=0.5
    )
    rabi3 = intensity_profile2.get_rabi_freq(pos1)
    assert rabi3[0] == 1
    rabi4 = intensity_profile2.get_rabi_freq(pos2)
    assert rabi4[0] == 0



"""Python package for simulating light-pulse atom interferometers."""

from importlib.metadata import version

from .atoms import AtomicEnsemble, create_random_ensemble
from .beam import IntensityProfile, Wavefront, Wavevectors, gen_wavefront
from .convert import cart2pol, phase_error_to_grav, pol2cart, temp, vel_from_temp
from .det import Detector, PolarDetector, SphericalDetector
from .dist import velocity_dist_for_box_pulse_velsel, velocity_dist_for_gaussian_velsel
from .prop import (
    FreePropagator,
    Propagator,
    SpatialSuperpositionTransitionPropagator,
    TwoLevelTransitionPropagator,
)
from .zern import ZernikeNorm, ZernikeOrder, ZernikePolynomial, j_to_n_m

__version__ = version("aisim")

__all__ = [
    "AtomicEnsemble",
    "create_random_ensemble",
    "IntensityProfile",
    "Wavefront",
    "Wavevectors",
    "gen_wavefront",
    "cart2pol",
    "phase_error_to_grav",
    "pol2cart",
    "temp",
    "vel_from_temp",
    "Detector",
    "PolarDetector",
    "SphericalDetector",
    "FreePropagator",
    "Propagator",
    "SpatialSuperpositionTransitionPropagator",
    "TwoLevelTransitionPropagator",
    "ZernikeNorm",
    "ZernikeOrder",
    "ZernikePolynomial",
    "j_to_n_m",
    "velocity_dist_for_box_pulse_velsel",
    "velocity_dist_for_gaussian_velsel",
]



"""Classes to propagate atomic ensembles."""

import copy

import numpy as np


class Propagator:
    """
    A generic propagator.

    This is just a template class without an implemented propagation matrix.

    Parameters
    ----------
    time_delta : float
        time that should be propagated
    **kwargs :
        Additional arguments used by classes that inherit from this class. All
        keyworded arguments are stored as attribues.
    """

    def __init__(self, time_delta, **kwargs):
        self.time_delta = time_delta

        # save all keyworded arguments as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def _prop_matrix(self, atoms):
        raise NotImplementedError()

    def propagate(self, atoms):
        """
        Propagate an atomic ensemble.

        Parameters
        ----------
        atoms : AtomicEnsemble
            atomic ensemble that should be is propagated
        """
        atoms = copy.deepcopy(atoms)
        # propagate position for half a time step
        atoms.position += atoms.velocity * self.time_delta / 2
        # U*|Psi>
        atoms.state_kets = np.einsum(
            "ijk,ikl ->ijl", self._prop_matrix(atoms), atoms.state_kets
        )
        # propagate for another half time step
        atoms.position += atoms.velocity * self.time_delta / 2
        # update internal time
        atoms.time += self.time_delta
        return atoms


class FreePropagator(Propagator):
    """
    Propagator implementing free propagation without light-matter interaction.

    Parameters
    ----------
    time_delta : float
        time that should be propagated
    """

    def _prop_matrix(self, atoms):
        n_levels = atoms.state_kets[0].shape[0]
        return np.repeat([np.eye(n_levels)], repeats=len(atoms), axis=0)


class TwoLevelTransitionPropagator(Propagator):
    """
    A time propagator of an effective Raman two-level system.

    Parameters
    ----------
    time_delta: float
        length of pulse
    intensity_profile : IntensityProfile
        Intensity profile of the interferometry lasers
    wave_vectors: Wavevectors
        wave vectors of the two Raman beams for calculation of Doppler shifts
    wf : Wavefront , optional
        wavefront aberrations of the interferometry beam
    phase_scan : float
        effective phase for fringe scans

    Notes
    -----
    The propagator is for example defined in  [1].

    References
    ----------
    [1] Young, B. C., Kasevich, M., & Chu, S. (1997). Precision atom interferometry with
    light pulses. In P. R. Berman (Ed.), Atom Interferometry (pp. 363–406). Academic
    Press.
    https://doi.org/10.1016/B978-012092460-8/50010-2
    """

    def __init__(
        self, time_delta, intensity_profile, wave_vectors=None, wf=None, phase_scan=0
    ):
        super().__init__(
            time_delta,
            intensity_profile=intensity_profile,
            wave_vectors=wave_vectors,
            wf=wf,
            phase_scan=phase_scan,
        )

    def _prop_matrix(self, atoms):
        # calculate the effective Rabi frequency at atoms' positions
        Omega_eff = self.intensity_profile.get_rabi_freq(atoms.position)
        if self.wf is None:
            phase = 0
        else:
            # calculate phase at atoms' positions
            phase = self.wf.get_value(atoms.position)

        phase += self.phase_scan

        if self.wave_vectors is None:
            delta = 0
        else:
            # calculate two photon detuning for atoms' velocity (-v*k_eff)
            delta = self.wave_vectors.doppler_shift(atoms)

        # calculate Rabi frequency
        Omega_R = np.sqrt(Omega_eff**2 + delta**2)
        # beginning of pulse t0
        t0 = atoms.time
        tau = self.time_delta

        # calculate matrix elements

        sin_theta = Omega_eff / Omega_R
        cos_theta = -delta / Omega_R

        u_ee = np.cos(Omega_R * tau / 2) - 1j * cos_theta * np.sin(Omega_R * tau / 2)
        u_ee *= np.exp(-1j * delta * tau / 2)

        u_eg = (
            np.exp(-1j * (delta * t0 + phase))
            * -1j
            * sin_theta
            * np.sin(Omega_R * tau / 2)
        )
        u_eg *= np.exp(-1j * delta * tau / 2)

        u_ge = (
            np.exp(+1j * (delta * t0 + phase))
            * -1j
            * sin_theta
            * np.sin(Omega_R * tau / 2)
        )
        u_ge *= np.exp(1j * delta * tau / 2)

        u_gg = np.cos(Omega_R * tau / 2) + 1j * cos_theta * np.sin(Omega_R * tau / 2)
        u_gg *= np.exp(1j * delta * tau / 2)

        u = np.array([[u_ee, u_eg], [u_ge, u_gg]], dtype="complex")
        u = np.transpose(u, (2, 0, 1))
        return u


class SpatialSuperpositionTransitionPropagator(TwoLevelTransitionPropagator):
    """
    An effective Raman two-level system.

    It is implemented as a time propagator as defined in [1]. In addition to
    class TwoLevelTransitionPropagator, this adds spatial superpositions
    in z-direction that occour at each pulse.

    Parameters
    ----------
    time_delta: float
        length of pulse
    intensity_profile : IntensityProfile
        Intensity profile of the interferometry lasers
    wave_vectors: Wavevectors
        wave vectors of the two Raman beams for calculation of Doppler shifts
    wf : Wavefront , optional
        wavefront aberrations of the interferometry beam
    phase_scan : float
        effective phase for fringe scans
    n_pulses : int
        overall number of intended light pulses in symmetric atom
        interferometry sequence. Each pulse adds two spatial eigenstates.
    n_pulse : int
        number of light pulse of symmetric atom-interferometry sequence.

    References
    ----------
    [1] Young, B. C., Kasevich, M., & Chu, S. (1997). Precision atom interferometry with
    light pulses. In P. R. Berman (Ed.), Atom Interferometry (pp. 363–406). Academic
    Press.
    https://doi.org/10.1016/B978-012092460-8/50010-2
    """

    def __init__(
        self,
        time_delta,
        intensity_profile,
        n_pulses,
        n_pulse,
        wave_vectors=None,
        wf=None,
        phase_scan=0,
    ):
        self.n_pulses = n_pulses
        self.n_pulse = n_pulse
        super().__init__(
            time_delta,
            intensity_profile=intensity_profile,
            wave_vectors=wave_vectors,
            wf=wf,
            phase_scan=phase_scan,
        )

    def _block_diag(self, u, num):
        """
        Fast generation of diagonal block matrices of internal state dynamics.

        Parameters
        ----------
        u : n × m × m array
            propagation matrices for the n atoms with m levels
        num : int
            number of times the internal propagation matrix is repeated in the diagonal
            block matrix

        Returns
        -------
        matrix : n × m*num × m*num
            block matrix mediating the interactions of the n atoms with m*num levels
        """
        # since u is a square matrix, row = cols = n_states
        n, m, m = u.shape
        matrix = np.zeros((n, num, m, num, m), dtype="complex")
        # Note that diag is a view of matrix, so matrix is changed. n is index for
        # atoms, m and k for the levels and i for num
        diag = np.einsum("nimik->nimk", matrix)
        for i in range(0, n):
            # diag[i, :] has shape (num, n_int, n_int), u[i] has shape (n_int, n_int),
            # i.e. so u[i] is copied num times in each loop
            diag[i, :] = u[i]
        # bring matrix into final form
        return matrix.reshape((n, m * num, m * num))

    def _index_shift(self):
        """
        Switch state indices to ensure that the correct states interact.

        Returns
        -------
        index_shift_matrix : 2*n_pulses × 2*n_pulses array
        """
        # Each pulse splits the initial state into two states (2 level system),
        # so we end up with 2*n_pulses output states
        index_shift_matrix = np.eye(2 * self.n_pulses)
        for i in range(0, len(index_shift_matrix)):
            if i % 2 == 0:
                # shift the index of one of the two interacting states;
                # applying this matrix n_pulses times results again in a unity matrx
                index_shift_matrix[i, :] = np.roll(index_shift_matrix[i, :], 2)
        return index_shift_matrix

    def _prop_matrix(self, atoms):
        assert atoms.state_kets.shape[1] == 2 * self.n_pulses, (
            "Number of states must be twice the number of pulses."
        )

        u_two_level = super()._prop_matrix(atoms)
        u = self._block_diag(u_two_level, self.n_pulses)
        # Create two index shift matrices. First, positions of the state vectors are
        # switched such that the two interacting states are next to each other and are
        # coupled by the blocks of _block_diag. After the application of the propagation
        # matrix, these states are put back into their initial position.
        shift_forth = np.linalg.matrix_power(self._index_shift(), self.n_pulse)
        shift_back = np.linalg.matrix_power(
            self._index_shift(), self.n_pulses - self.n_pulse
        )
        return np.einsum("ij,njk,kl->nil", shift_back, u, shift_forth)



"""Functions to convert various quantities."""

import numpy as np

mass = {"Rb87": 1.443161e-25}
"""Atomic mass in kg."""

kb = 1.3806488e-23
"""Boltzmann constant in J/K."""


def temp(sigma_v, species="Rb87"):
    """
    Calculate the temperature of an atomic cloud from its velocity spread.

    Parameters
    ----------
    sigma_v : float
        velocity spread (1 sigma) in meters per second
    species : str
        the atomic species, has to be a key in `mass`

    Returns
    -------
    temp : float
        temperature of the cloud in Kelvin

    Raises
    ------
    ValueError
        If negative velocity spread is passed
    """
    if sigma_v < 0:
        raise ValueError("sigma_v must be non-negative")
    return mass[species] * sigma_v**2 / kb


def vel_from_temp(temp, species="Rb87"):
    """
    Calculate the velocity spread (1 sigma) from the temperature of the cloud.

    Parameters
    ----------
    temp : float
        temperature of the cloud in Kelvin
    species : str
        the atomic species, has to be a key in `mass`

    Returns
    -------
    vel : float
        velocity spread (1 sigma)

    Raises
    ------
    ValueError
        If negative temperature is passed
    """
    if temp < 0:
        raise ValueError("temp must be non-negative")
    return np.sqrt(temp * kb / mass[species])


def cart2pol(cart):
    """
    Convert vectors in cartesian coordinates to polar coordinates.

    Parameters
    ----------
    cart : n × 3 array
        array of n vectors in cartesian coordinates (x, y, z)

    Returns
    -------
    pol : n × 3 array
        array of n vectors in polar coordinates (rho, theta, z)
    """
    x = cart[:, 0]
    y = cart[:, 1]
    z = cart[:, 2]
    rho = np.sqrt(x**2 + y**2)
    theta = np.arctan2(y, x)
    # Transpose to retain form (n × 3 array)
    return np.array([rho, theta, z]).T


def pol2cart(pol):
    """
    Convert vectors in polar coordinates to cartesian coordinates.

    Parameters
    ----------
    pol : n × 3 array
        array of n vectors in polar coordinates (rho, theta, z)

    Returns
    -------
    cart : n × 3 array
        array of n vectors in cartesian coordinates (x, y, z)
    """
    rho = pol[:, 0]
    theta = pol[:, 1]
    z = pol[:, 2]
    x = rho * np.cos(theta)
    y = rho * np.sin(theta)
    # Transpose to retain form (n × 3 array)
    return np.array([x, y, z]).T


def phase_error_to_grav(phase, T, keff):
    """
    Convert a phase error to gravitational acceleration.

    Takes the phase shift measured in a Mach Zehnder atom interferometer and converts it
    to the corresponding gravitional accleration.

    Parameters
    ----------
    phase : float
        Interferometer phase in rad
    T : float
        interferometer time in s
    keff : float
        effective wavenumber in rad/m

    Returns
    -------
    float :
        gravitational acceleration in  in m/s^2
    """
    return phase / keff / (T**2)


def arrival_time(z, t0=0.0, z0=0.0, v0=0.0, g=9.80665):
    """
    Calculate time when the atomic ensemble reaches a certaini position.

    Parameters
    ----------
    z : float
        position in m
    t0 : float
        time reference in s (at which z0 and v0 are known)
    z0, v0 : float
        initial position and velocity (in m and m/s, respectively), i.e. position and
        velocity at t_ref
    g : float
        gravitational acceleration in m/s**2

    Returns
    -------
    t : list of float
        The two times when atoms reach the position z in seconds.
    """
    t12 = [
        (v0 - np.sqrt(v0**2 + 2 * g * (z0 - z))) / g + t0,
        (v0 + np.sqrt(v0**2 + 2 * g * (z0 - z))) / g + t0,
    ]
    return t12



from enum import StrEnum
from functools import partial
from typing import Callable

import numpy as np
from scipy.special import eval_jacobi


# FIXME: This is not a very elegant solution. IT would be better to find an explicit
# formula for the conversion from Wyant/Fringe single index j to n, m rather than
# populating a lookup table.
def _n_m_to_j_wyant(n, m):
    return int((1 + (n + np.abs(m)) / (2)) ** 2 - 2 * np.abs(m) + (1 - np.sign(m)) / 2)


_wyant_j_to_n_m_lookup_table = {}

for n in range(100):
    for m in reversed(range(-n, n + 1)):
        if (n - abs(m)) % 2 == 0:
            j = int(_n_m_to_j_wyant(n, m))
            _wyant_j_to_n_m_lookup_table[j - 1] = (n, m)
_wyant_j_to_n_m_lookup_table = dict(sorted(_wyant_j_to_n_m_lookup_table.items()))
_fringe_j_to_n_m_lookup_table = {
    idx + 1: val for idx, val in _wyant_j_to_n_m_lookup_table.items()
}


class ZernikeOrder(StrEnum):
    """Enumeration of conventions for ordering Zernike polynomials with a single index.

    There are different ways to order the Zernike polynomials with a single index. The
    most common conventions are the OSA or ANSI convention [1, 2], the Noll
    convention [3], the Fringe/University of Arizona/Air Force convention [4],
    and the Wyant convention [5].

    References
    ----------
    [1] Noll, R. J. (1976).
        Zernike polynomials and atmospheric turbulence*.
        J. Opt. Soc. Am., 66, 207–211.
        https://opg.optica.org/josa/fulltext.cfm?uri=josa-66-3-207&id=56041
    [2] Thibos, L. N., Applegate, R. A., Schwiegerling, J. T., & Webb, R. (2000).
        Standards for Reporting the Optical Aberrations of Eyes.
        Optics InfoBase Conference Papers.
    [3] ANSI Z80.28-2017.
        American National Standard for Ophthalmics - Methods for Reporting Optical
        Aberrations Of Eyes.
    [4] Yen, A. (2021).
        Straightforward path to Zernike polynomials.
        Journal of Micro/Nanopatterning, Materials and Metrology, 20(2).
        https://doi.org/10.1117/1.JMM.20.2.020501
    [5] Wikipedia contributors. (2025).
        Zernike polynomials.
        In Wikipedia, The Free Encyclopedia.
        https://en.wikipedia.org/wiki/Zernike_polynomials
    """

    ANSI = "ANSI"
    """Ordering according to OSA and ANSI Z80.28-2017."""
    NOLL = "NOLL"
    """Noll ordering."""
    FRINGE = "FRINGE"
    """Fringe/University of Arizona/Air Force  ordering."""
    WYANT = "WYANT"
    """Wyant's ordering."""


FIRST_INDEX_J: dict[ZernikeOrder, int] = {
    ZernikeOrder.ANSI: 0,
    ZernikeOrder.NOLL: 1,
    ZernikeOrder.FRINGE: 1,
    ZernikeOrder.WYANT: 0,
}
"""First index j of the Zernike polynomials for each ordering scheme (1 or 0)."""


class ZernikeNorm(StrEnum):
    """Enumeration of Zernike conventions for normalization.

    There are two common conventions for normalizing Zernike polynomials. The Noll
    convention [1] normalizes the polynomials with a factor sqrt(n+1), while the OSA
    convention [2, 3] normalizes the polynomials to have unity variance.

    References
    ----------
    [1] Noll, R. J. (1976).
        Zernike polynomials and atmospheric turbulence*.
        J. Opt. Soc. Am., 66, 207-211.
        https://opg.optica.org/josa/fulltext.cfm?uri=josa-66-3-207&id=56041
    [2] Thibos, L. N., Applegate, R. A., Schwiegerling, J. T., & Webb, R. (2000).
        Standards for Reporting the Optical Aberrations of Eyes.
        Optics InfoBase Conference Papers.
        https://doi.org/10.3928/1081-597x-20020901-30
    [3] Lakshminarayanan, V., & Fleck, A. (2011).
        Zernike polynomials: A guide.
        Journal of Modern Optics (Vol. 58, Issue 7).
        https://doi.org/10.1080/09500340.2011.554896
    """

    NOLL = "NOLL"
    """Noll normalization [1], sqrt(n+1)."""
    OSA = "OSA"
    """OSA normalization [2], sqrt(2n+2) for m=0 and sqrt(n+1) for m!=0."""


def j_to_n_m(j: int, order: ZernikeOrder) -> tuple[int, int]:
    """Map the single index j to the pair of indices (n, m).

    Parameters
    ----------
    j : int
        Single index of the Zernike polynomial.
    order : ZernikeOrder
        Ordering scheme for the Zernike polynomials.

    Returns
    -------
    tuple of int
        Degree and azimuth of the Zernike polynomial.

    Notes
    -----
    Some of the implementations were taken from the hcipy library:
    https://github.com/ehpor/hcipy/blob/master/hcipy/mode_basis/zernike.py
    """
    if order not in FIRST_INDEX_J:
        raise ValueError(f"Unknown ordering scheme {order=}.")
    if j < FIRST_INDEX_J[order]:
        raise ValueError(
            f"Index {j=} must be greater or equal to {FIRST_INDEX_J[order]} for {order}."
        )
    match order:
        case ZernikeOrder.NOLL:
            n = int(np.sqrt(2 * j - 1) + 0.5) - 1
            if n % 2:
                m = 2 * int((2 * (j + 1) - n * (n + 1)) // 4) - 1
            else:
                m = 2 * int((2 * j + 1 - n * (n + 1)) // 4)
            m = m * (-1) ** (j % 2)
        case ZernikeOrder.ANSI:
            n = int((np.sqrt(8 * j + 1) - 1) / 2)
            m = 2 * j - n * (n + 2)
        case ZernikeOrder.WYANT:
            if j >= len(_wyant_j_to_n_m_lookup_table):
                raise ValueError(f"Fringe index {j=} is out of range.")
            n, m = _wyant_j_to_n_m_lookup_table[j]
        case ZernikeOrder.FRINGE:
            if j > len(_fringe_j_to_n_m_lookup_table):
                raise ValueError(f"Fringe index {j=} is out of range.")
            n, m = _fringe_j_to_n_m_lookup_table[j]
    return n, m


def radial(rho: np.ndarray, n: int, m: int) -> np.ndarray:
    """Compute the radial part of the Zernike polynomial.

    Parameters
    ----------
    rho : np.ndarray
        Normalized adial coordinate.
    n, m : int
        Degree and azimuth of the Zernike polynomial.

    Returns
    -------
    np.ndarray
        Value of the Zernike polynomial at the given coordinates.
    """
    m = np.abs(m)
    r = (
        (-1) ** ((n - m) // 2)
        * rho**m
        * eval_jacobi((n - m) // 2, m, 0, 1 - 2 * rho**2)
    )
    r[np.isnan(rho)] = np.nan
    return r


def zernike_term(
    rho: np.ndarray, theta: np.ndarray, n: int, m: int, norm: ZernikeNorm | None = None
) -> np.ndarray:
    """Compute the Zernike polynomial of degree n and azimuth m.

    rho : np.ndarray
        Normalized radial coordinate.
    theta : np.ndarray
        Azimuthal coordinate.
    n, m : int
        Degree and azimuth of the Zernike polynomial.
    norm : ZernikeNorm, optional
        Normalization scheme for the Zernike polynomial. Default is None, which means
        no normalization.

    Returns
    -------
    np.ndarray
        Value of the Zernike polynomial at the given coordinates.
    """
    match norm:
        case None:
            norm = 1
        case ZernikeNorm.NOLL:
            norm = np.sqrt(n + 1)
        case ZernikeNorm.OSA:
            norm = np.sqrt((2 * n + 2) / (1 + int(m == 0)))
        case _:
            raise ValueError(f"Unknown normalization scheme {norm=}.")
    r = radial(rho, n, m)
    r = r * norm
    if m == 0:
        return r
    elif m < 0:
        return r * np.sin(np.abs(m) * theta)
    else:
        return r * np.cos(m * theta)


class ZernikePolynomial:
    """Class for evaluating Zernike polynomials.

    Parameters
    ----------
    coeffs : dict
        Coefficients of the Zernike polynomial. The keys of the dict are the single
        indices j of the Zernike polynomials according to the given ordering scheme. If
        the ordering scheme starts at 1, index 0 (if present) is ignored.
    order : ZernikeOrder or str
        Ordering scheme for the Zernike polynomials. Default is ZernikeOrder.NOLL. See
        `ZernikeOrder` for possible values.
    norm : ZernikeNorm, str or None (optional)
        Normalization scheme for the Zernike polynomials. Default is None, which means
        no normalization. See `ZernikeNorm` for possible values.

    Attributes
    ----------
    coeffs : dict
        Coefficients of the Zernike polynomial with the single index j as keys.
    order : ZernikeOrder or str
        Ordering scheme for the Zernike polynomials. See `ZernikeOrder` for possible
        values.
    norm : ZernikeNorm, str or None
        Normalization scheme for the Zernike polynomials. See `ZernikeNorm` for possible
        values.
    zernike_funcs : dict
        Dict of Zernike polynomial functions for each coefficient indexed by  the single
        index j.
    """

    def __init__(
        self,
        coeffs: dict[int, float],
        order: ZernikeOrder = ZernikeOrder.NOLL,
        norm: ZernikeNorm | None = None,
    ) -> None:
        self.coeffs = coeffs
        self.order = order
        self.norm = norm
        self.zernike_funcs: dict[int, Callable] = {}
        for j in self.coeffs:
            n, m = j_to_n_m(j, order)
            self.zernike_funcs[j] = partial(zernike_term, n=n, m=m, norm=self.norm)

    def zern_terms(self, rho: np.ndarray, theta: np.ndarray) -> dict[int, np.ndarray]:
        """Evaluate the Zernike polynomial terms at the given coordinates.

        Parameters
        ----------
        rho :  ndarray
            1 dimensional array of normalized radial coordinates.
        theta : ndarray
            1 dimensional array of azimuthal coordinates. Must have the same length as
            `rho`.

        Returns
        -------
        dict
            Dictionary with the Zernike polynomial terms for all coordinates and for
            each coefficient. The keys are the single indices j depending on the
            ordering scheme.
        """
        terms = {}
        for j in self.coeffs:
            terms[j] = self.coeffs[j] * self.zernike_funcs[j](rho, theta)
        return terms

    def zern_sum(self, rho: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """Calculate the sum of the Zernike polynomials at the given coordinates.

        Parameters
        ----------
        rho :  ndarray
            1 dimensional array of normalized radial coordinates.
        theta : ndarray
            1 dimensional array of azimuthal coordinates. Must have the same length as
            `rho`.

        Returns
        -------
        ndarray
            n dimensional array of the sum of the Zernike polynomial terms at the n
            coordinates.
        """
        return np.sum(np.array(list(self.zern_terms(rho, theta).values())), axis=0)



import numpy as np
import pytest

import aisim as ais


def test_mass():
    for mass in ais.convert.mass.values():
        assert mass > 0


def test_temp():
    for species in ais.convert.mass:
        assert ais.convert.temp(0, species=species) == 0
        assert ais.convert.temp(1.0, species=species) > 0
        with pytest.raises(ValueError):
            ais.convert.temp(-1.0, species=species)


def test_vel_from_temp():
    for species in ais.convert.mass:
        assert ais.convert.vel_from_temp(0, species=species) == 0
        assert ais.convert.vel_from_temp(1.0, species=species) > 0
        with pytest.raises(ValueError):
            ais.convert.vel_from_temp(-1.0, species=species)


def test_pol2cart_and_cart2pol():
    cart = np.array([[0, 0, 0], [1, 2, 3], [-1, -2, -3], [-10, 10, -10]])
    pol = ais.convert.cart2pol(cart)
    np.testing.assert_array_almost_equal(cart, ais.convert.pol2cart(pol))


def test_arrival_time():
    t12 = ais.convert.arrival_time(-1, 0, 0, 0, 1)
    assert t12[0] == -np.sqrt(2)
    assert t12[1] == np.sqrt(2)
